from .wiktionary import WiktionaryDefinitionCorpus
from .codwoe import CODWOEDataSet
from .cpae import CPAEDataSet
from .allnli import AllNLIDataSet
from .stsb import STSBDataSet
from .entailmentbank import EntailmentBankDataSet
from .wordnet_filtered import WordNetFilteredDataSet
from .wordnet_spanish import SpanishWordNetFilteredDataSet
from .inference_types import InferenceTypesDataSet
